package ch.psi.daq.queryrest.controller.validator;

import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.Set;

import javax.annotation.Resource;

import org.springframework.validation.Errors;
import org.springframework.validation.Validator;

import ch.psi.daq.query.model.Aggregation;
import ch.psi.daq.query.model.DBMode;
import ch.psi.daq.query.model.QueryField;
import ch.psi.daq.query.model.impl.DAQQuery;
import ch.psi.daq.cassandra.request.Request;
import ch.psi.daq.queryrest.config.QueryRestConfig;

public class QueryValidator implements Validator {

   @Resource(name = QueryRestConfig.BEAN_NAME_DEFAULT_RESPONSE_FIELDS)
   private Set<QueryField> defaultResponseFields;

   @Resource(name = QueryRestConfig.BEAN_NAME_DEFAULT_RESPONSE_AGGREGATIONS)
   private Set<Aggregation> defaultResponseAggregations;

   /**
    * {@inheritDoc}
    */
   @Override
   public boolean supports(Class<?> clazz) {   
      return DAQQuery.class.isAssignableFrom(clazz);
   }

   /**
    * {@inheritDoc}
    */
   @Override
   public void validate(Object target, Errors errors) {

      DAQQuery query = (DAQQuery) target;
      
      Request request = query.getRequest();

      if (DBMode.archiverappliance.equals(query.getDbMode())) {
         if (!request.getRequestRange().isTimeRangeDefined()) {
            errors.reject("dbMode", "ArchiverAppliance supports time range queries only!");
         }
      }

      // set default values (if not set)
      if (query.getFields() == null || query.getFields().isEmpty()) {
         query.setFields(new LinkedHashSet<>(defaultResponseFields));
      }

      if (query.getAggregations() == null || query.getAggregations().isEmpty()) {
         query.setAggregations(new LinkedList<>(defaultResponseAggregations));
      }
   }

}
