package ch.psi.daq.queryrest.filter;

import java.io.IOException;
import java.util.Arrays;
import java.util.Set;
import java.util.stream.Collectors;

import javax.annotation.PostConstruct;
import javax.annotation.Resource;
import javax.servlet.FilterChain;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.springframework.web.filter.OncePerRequestFilter;

import ch.psi.daq.queryrest.config.QueryRestConfig;

public class CorsFilter extends OncePerRequestFilter {

   private static final String ALLOW_ORIGIN_HEADER = "Access-Control-Allow-Origin";

   @Resource(name = QueryRestConfig.BEAN_NAME_CORS_ALLOWEDORIGINS)
   private String configuredOrigins;
   
   @Resource(name = QueryRestConfig.BEAN_NAME_CORS_FORCEALLHEADERS)
   private Boolean forceAllHeaders;
   
   private Set<String> allowedOrigins;
   
   @PostConstruct
   public void afterPropertiesSet(){     
      allowedOrigins = Arrays.stream(configuredOrigins.split(","))
            .map(s -> s.trim())
            .collect(Collectors.toSet());
   }

   @Override
   protected void doFilterInternal(HttpServletRequest request, HttpServletResponse response, FilterChain filterChain)
         throws ServletException, IOException {     
      String originHeader = request.getHeader("Origin");
      if (forceAllHeaders) {
         // include headers no matter what - good for development
         if (allowedOrigins.contains(originHeader)) {
            response.addHeader(ALLOW_ORIGIN_HEADER, originHeader);
         } else {
            response.addHeader(ALLOW_ORIGIN_HEADER, "*");
         }
         setDefaultCorsHeaders(response);
         
      } else if (request.getHeader("Access-Control-Request-Method") != null && "OPTIONS".equals(request.getMethod())) {
         // this is for 'real' Cross-site browser requests
         if (allowedOrigins.contains(originHeader)) {
            response.addHeader(ALLOW_ORIGIN_HEADER, originHeader);
         }
         setDefaultCorsHeaders(response);
      }

      filterChain.doFilter(request, response);
   }


   private void setDefaultCorsHeaders(HttpServletResponse response) {
      response.addHeader("Access-Control-Allow-Methods", "GET, POST, PUT, DELETE, OPTIONS");
      response.addHeader("Access-Control-Allow-Headers", "Origin, Authorization, Accept, Content-Type");
      response.addHeader("Access-Control-Max-Age", "1800");
   }


   public void setConfiguredOrigins(String configuredOrigins) {
      this.configuredOrigins = configuredOrigins;
   }


   public void setForceAllHeaders(boolean forceAllHeaders) {
      this.forceAllHeaders = forceAllHeaders;
   }
   
}
