package ch.psi.daq.test.queryrest;

import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.ComponentScan;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Import;
import org.springframework.context.annotation.Lazy;
import org.springframework.context.annotation.PropertySource;
import org.springframework.context.annotation.PropertySources;
import org.springframework.web.servlet.config.annotation.EnableWebMvc;
import org.springframework.web.servlet.config.annotation.WebMvcConfigurationSupport;

import ch.psi.daq.cassandra.reader.CassandraReader;
import ch.psi.daq.cassandra.util.test.CassandraDataGen;
import ch.psi.daq.query.processor.QueryProcessor;
import ch.psi.daq.query.processor.QueryProcessorLocal;
import ch.psi.daq.test.cassandra.admin.CassandraTestAdmin;
import ch.psi.daq.test.cassandra.admin.CassandraTestAdminImpl;
import ch.psi.daq.test.query.config.LocalQueryTestConfig;
import ch.psi.daq.test.queryrest.query.DummyCassandraReader;

@Configuration
@ComponentScan
@EnableWebMvc
@PropertySources(value = {
      @PropertySource(value = {"classpath:queryrest-test.properties"})
})
public class DaqWebMvcConfig extends WebMvcConfigurationSupport {

   // ensure that properties in dispatcher.properties are loaded first and then overwritten by the
   // properties in dispatcher-test.properties
   @Import(value = {LocalQueryTestConfig.class})
   static class InnerConfiguration {
   }
   
   @Bean
   @Lazy
   public QueryProcessor cassandraQueryProcessorLocal() {
      return new QueryProcessorLocal(cassandraReader());
   }
   
   @Bean
   @Lazy
   CassandraReader cassandraReader() {
      return new DummyCassandraReader();
   }
   
   @Bean
   @Lazy
   public CassandraTestAdmin cassandraTestAdmin() {
      return new CassandraTestAdminImpl();
   }
   
   @Bean
   @Lazy
   public CassandraDataGen cassandraDataGen() {
      return new CassandraDataGen();
   }
}